"""Set of tests for feature_selector module."""

from pytest import fixture

from AutoCarver.feature_selector import *


def test_feature_selector_quanti():
    """TODO"""
    pass
    # assert False, "TODO"


def test_feature_selector_quali():
    """TODO"""
    pass
    # assert False, "TODO"
